package com.jprofiler.api.agent.probe;

/**
 * Defines a single method that should be intercepted by an @{link InterceptorProbe interceptor probe}.
 * You can define a method in a concrete class or a wildcard interception that intercepts all methods based on the
 * method name an signature only. For wildcard interceptions, make sure to try to intercept uncommon method signatures
 * only in order to reduce the instrumentation overhead. For example, intercepting <tt>getName()</tt> as a wildcard
 * interception leads to many unwanted instrumentations.
 * <p>
 * The method signature is written in JVM format. This format is also used by JNI and is of the form:
 * <pre>
 * (argument-type*)return-type
 * </pre>
 * and uses the JVM's representation of type signatures for both argument type and return type. The following table
 * shows all type signatures:
 * <p><table border="1"><tr><th>Type</th><th>Signature</th></tr>
 * <tr><td>Z</td><td>boolean</td></tr>
 * <tr><td>B</td><td>byte</td></tr>
 * <tr><td>C</td><td>char</td></tr>
 * <tr><td>S</td><td>short</td></tr>
 * <tr><td>I</td><td>int</td></tr>
 * <tr><td>J</td><td>long</td></tr>
 * <tr><td>F</td><td>float</td></tr>
 * <tr><td>D</td><td>double</td></tr>
 * <tr><td>L fully-qualified-class ;</td><td>fully-qualified-class</td></tr> 
 * <tr><td>[ type</td><td>type[]</td></tr>
 * </table>
 * <p>
 * The fully qualified class references have to be written with '/' package separators instead of the usual '.'
 * package separators. There must be no spaces in a type signature.
 * <p>
 * For example, the Java method:
 * <pre>
 * long someMethod (int n, String s, int[] arr);
 * </pre>
 * has the following type signature:
 * <pre>
 * (ILjava/lang/String;[I)J
 * </pre>
 *
 * @see InterceptorProbe
 */
public class InterceptionMethod {

    private String className;
    private String methodName;
    private String methodSignature;

    /**
     * Constructor for a method interception of a concrete class.
     *
     * @param className the fully qualified name of the class that contains the intercepted method. Can be written
     * with '/' or '.' package separators. If <tt>null</tt>, this constructor is equivalent to {@link #InterceptionMethod(String, String)}.
     * @param methodName the name of the intercepted method
     * @param methodSignature the signature of the intercepted method in the JVM format explained above
     */
    public InterceptionMethod(String className, String methodName, String methodSignature) {
        
        if (className != null) {
            className = className.replace('.', '/');
        }
        this.className = className;
        this.methodName = methodName;
        this.methodSignature = methodSignature;
    }

    /**
     * Constructor for a wildcard method interception. All classes are checked whether they contains such a method
     * and are instrumented accordingly.
     *
     * @param methodName the name of the intercepted method
     * @param methodSignature the signature of the intercepted method in the JVM format explained above
     */
    public InterceptionMethod(String methodName, String methodSignature) {
        this(null, methodName, methodSignature);
    }

    /**
     * Returns the class name.
     * @return the class name
     */
    public String getClassName() {
        return className;
    }

    /**
     * Returns the method name.
     * @return the method name
     */
    public String getMethodName() {
        return methodName;
    }

    /**
     * Returns the method signature
     * @return the method signature in the format explained above.
     */
    public String getMethodSignature() {
        return methodSignature;
    }

    public String toString() {
        return "InterceptionMethod{" +
                "className='" + className + '\'' +
                ", methodName='" + methodName + '\'' +
                ", methodSignature='" + methodSignature + '\'' +
                '}';
    }
}
