package com.jprofiler.api.agent.probe;

/**
 * Context object that is passed to the interception methods of {@link InterceptorProbe interceptor probes}.
 * It extends the capabilities of the {@link ProbeContext parent interface} to provide support for
 * payload creation and management.
 * <p>
 *     The <b>help topic on custom probes</b> in the bundled documentation explains how to develop custom probes.
 * </p>
 */
public interface InterceptorContext extends ProbeContext {

    /**
     * Creates a payload info object for this interceptor probe. See {@link PayloadInfo} for more information.
     * @param description the payload description
     * @return the payload info object
     */
    PayloadInfo createPayloadInfo(String description);

    /**
     * Creates a payload info object for this interceptor probe with a memento object. See {@link PayloadInfo} for more information.
     * @param description the payload description
     * @param memento a memento object for information that you need in order to analyze the payload info in a different context.
     * @return the payload info object
     */
    PayloadInfo createPayloadInfo(String description, Object memento);

    /**
     * Helper method to check whether the class of an object implements an interface. This method uses reflection
     * to check the condition and saves the result in a thread-local cache, so subsequent invocations are very fast.
     * @param object the object whose class should be checked
     * @param interfaceName the name of the interface that should be checked
     * @return if the interface is implemented or not
     */
    boolean implementsInterface(Object object, String interfaceName);

    /**
     * Helper method to check if a method name and signature are equivalent with target values. This method is important if
     * you intercept multiple methods, since all method calls are intercepted by the same methods in the interceptor.
     * @param actualMethodName the method name as passed to the interceptor methods
     * @param actualMethodSignature the method signature as passed to the interceptor methods
     * @param targetMethodName the method name against which you are testing
     * @param targetMethodSignature the method signature against which you are testing in JVM format. See {@link InterceptionMethod} for more information on that format.
     * @return if the methods are equal or not
     */
    boolean isTargetMethod(String actualMethodName, String actualMethodSignature, String targetMethodName, String targetMethodSignature);

    /**
     * Publishes the payload into the call tree. You can see the results of published payload in the hot spot view.
     * The probe meta data method {@link ProbeMetaData#payload(boolean)} has to be called with the argument <tt>true</tt>
     * in the {@link com.jprofiler.api.agent.probe.Probe#getMetaData() meta data configuration method}.
     * @param payloadInfo the payload that should be published.
     */
    void addPayloadInfo(PayloadInfo payloadInfo);

    /**
     * Saves a payload to a thread local stack. You can retrieve it with
     * {@link #pop()} later on. The typical use case is to create a payload during the method entry, creating its
     * payload name from the arguments of the intercepted method or other data. During method exit, you then calculate
     * times and publish it.
     * @param payloadInfo the payload info
     */
    void push(PayloadInfo payloadInfo);

    /**
     * Retrieves a payload that was saved to the thread local stack with {@link #push(PayloadInfo)}.
     * @return the most recently saved payload info. If the payload stack is empty, it returns <tt>null</tt>.
     */
    PayloadInfo pop();

    /**
     * Check if the payload stack is empty. Payloads are saved to the payload stack with {@link #push(PayloadInfo)}.
     * @return the value
     */
    public boolean isPayloadStackEmpty();

    /**
     * Creates an event for opening a control object with payload information. This method enhances the
     * {@link ProbeContext#createOpenEvent(String, Object)} with payload. The description of the event is taken from the
     * payload name.
     * <p>
     *     Note that the returned event must be published with {@link #addEvent(ProbeEvent)}, otherwise
     *     the event will be discarded.
     * </p>
     * @param payloadInfo the payload info
     * @param controlObject the control object that is being opened
     * @return the probe event
     */
    ProbeEvent createOpenEvent(PayloadInfo payloadInfo, Object controlObject);

    /**
     * Creates an event for closing a control object with payload information. This method enhances the
     * {@link ProbeContext#createCloseEvent(String, Object)} with payload. The description of the event is taken from the
     * payload name.
     * <p>
     *     Note that the returned event must be published with {@link #addEvent(ProbeEvent)}, otherwise
     *     the event will be discarded.
     * </p>
     * @param payloadInfo the payload info
     * @param controlObject the control object that is being closed
     * @return the probe event
     */
    ProbeEvent createCloseEvent(PayloadInfo payloadInfo, Object controlObject);

    /**
     * Creates a custom event with payload information. This method enhances the
     * {@link ProbeContext#createCustomEvent(String, int, Object)}. The description of the event is taken from the
     * payload name.
     * <p>
     *     Note that the returned event must be published with {@link #addEvent(ProbeEvent)}, otherwise
     *     the event will be discarded.
     * </p>
     * @param payloadInfo the payload info
     * @param type the type corresponds to the index of the desired custom event in the the call to
     * {@link ProbeMetaData#customTypeNames(String[])}
     * @param controlObject the control object for this event. Can be <tt>null</tt>.
     * @return the probe event
     */
    ProbeEvent createCustomEvent(PayloadInfo payloadInfo, int type, Object controlObject);

    /**
     * Change the state of a control object for the time line view.
     * In the time line view, the horizontal bars that show the control objects on the time axis
     * are colored according to the colors defined in
     * {@link ProbeMetaData#customColors(int[])}. The indices correspond to the custom states defined with
     * {@link ProbeMetaData#customTypeNames(String[])}. Usually, this is done automatically based on the published
     * custom events. If you do not want to publish, associated events, you can used this method to change the state of
     * a control object.
     * @param controlObject the control object whose state should be changed
     * @param type the new state. This is the index of the desired custom state in the call to
     * {@link ProbeMetaData#customTypeNames(String[])}.
     */
    void startCustomState(Object controlObject, int type);

    /**
     * Returns the description that was used for creating a control object. This corresponds to the
     * <tt>description</tt> parameter in {@link #createOpenEvent(PayloadInfo, Object)}.
     * @param controlObject the control object
     * @return the description
     */
    String getOpenEventDescription(Object controlObject);
}
